/*
 * $Id: arch-solaris-distrib.h,v 1.1.2.5 1997/12/10 02:10:02 randall Exp $
 */
/*
 * $Log: arch-solaris-distrib.h,v $
 * Revision 1.1.2.5  1997/12/10 02:10:02  randall
 * Some small changes.  Working on fixing TLB shootdown code.
 *
 * Revision 1.1.2.4  1997/11/13 18:51:13  randall
 * Update of solaris UDP version.
 *
 * Revision 1.1.2.3  1997/10/14 20:47:05  randall
 * Fixed to make sure distributed version still works on xolas.
 *
 * Revision 1.1.2.2  1997/10/14 20:25:41  randall
 * First working version of distributed cilk on DEC machines.
 *
 * Revision 1.1.2.1  1997/06/13 22:33:04  randall
 * Added distributed version to the repository under
 * branch "distributed"
 *
 * Revision 1.5  1996/12/10 21:41:54  athena
 * Massive renaming of variables to ease port to architechtures
 * whose address space is not so flexible as on Solaris.
 * All global variables are now either SHARED, PRIVATE or SHARED-ReadOnly.
 * All accesses are wrapped with macros, and (almost) all variables
 * are initialized.  The net effect is that it should be easy to
 * put all shared variables in a structure, and things like that.
 *
 * Revision 1.4  1996/12/04 18:55:23  athena
 * Changed memory locks to use CAS as opposed to SWAP (SWAP
 * is deprecated in V9)
 *
 * Revision 1.3  1996/12/04 00:29:48  athena
 * Fixed wall-clock timer in cilk.c
 *
 * Revision 1.2  1996/12/03 23:09:02  athena
 * Removed all old Cilk-4 process/wall timers (in favor of
 * microstate acocunting)
 *
 * Revision 1.1.1.1  1996/12/03 22:20:57  athena
 * Created repository for Cilk 5 (a.k.a. Milk)
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef SPARC_V9
#define SPARC_V9 0
#endif

#ifndef SPARC_V8
#define SPARC_V8 (!(SPARC_V9))
#endif

#define CACHE_LINE 64

/* padding, to be used in structures to avoid false sharing */
#define CILK_CACHE_LINE_PAD  char __dummy[CACHE_LINE]

#define CACHE_LINE_ALIGNED(decl) decl  __attribute__((aligned(CACHE_LINE)))

/* safety test */
#if SPARC_V8 && SPARC_V9
#error define either SPARC_V8 or SPARC_V9, but not both
#endif

#if (!SPARC_V8) && (!SPARC_V9)
#error define either SPARC_V8 or SPARC_V9
#endif

#if (!CILK_SOLARIS_DISTRIB)
#error CILK_SOLARIS_DISTRIB not defined !
#endif

#define CILK_USE_MEMORY_LOCKS 1

/***********************************************************\
 * Lock implementation
\***********************************************************/

#if CILK_USE_MEMORY_LOCKS
/* a lock variable */
typedef volatile int Cilk_mutex;

#else
/* use solaris locks */
#include <thread.h>
#include <synch.h>

typedef volatile mutex_t Cilk_mutex;
#endif

/***********************************************************\
 * Machine size
\***********************************************************/

#include <unistd.h>

static inline int Cilk_partition_size(void)
{
     /* 
      * This is only really right for an unloaded machine.
      */
     return sysconf(_SC_NPROCESSORS_ONLN);
}

/***********************************************************
 * Interaction with unix
 ***********************************************************/
#define Cilk_yield()     { yield(); }

/***********************************************************\
 * Various types of membars
\***********************************************************/

#if SPARC_V8
/* helper function - a hardware swap instruction */
static inline int Cilk_swap(register volatile int *ptr, register int xchg)
{
     asm volatile ("swap [%1],%0"
		   :"=r" (xchg)	/* registers written */
		   :"r"(ptr), "0"(xchg)		/* registers read */
     );
     return xchg;
}
#endif

#if SPARC_V9
/*
 * Use of SWAP is deprecated on V9 (see manual, A.56).
 * The manual suggests to implement locks with CAS, rather than
 * using SWAP.
 *
 * CAS (Compare And Swap) works as follows: CAS r0, r1, r2
 * reads [r0], compares with r1, if they are the same swaps
 * [r0] and r2, otherwise loads r2 with [r0] and leaves [r0]
 * unchanged.  Everything happens atomically, of course.
 * Also, according to the manual (8.4.4.3), "atomic load-stores
 * are ordered with respect to load and stores", which should
 * ensure the release consistency we want.
 *
 * Experimentally, CAS seems to be a bit faster than SWAP.
 * I don't know if this happens because CAS saves the extra store
 * (sometimes).
 */
static inline int Cilk_cas(register volatile int *ptr, register int cmp,
			   register int xchg)
{
     asm volatile ("cas [%1],%3,%0"
		   :"=r" (xchg)	/* registers written */
		   :"r"(ptr), "0"(xchg), "r"(cmp) /* registers read */
     );
     return xchg;
}
#endif

/*
 *  Ensure that all previous memory operations are completed before
 *  continuing.
 */
static inline void Cilk_fence(void)
{
#if SPARC_V9
     asm volatile ("membar #StoreLoad | #StoreStore | #LoadLoad | #LoadStore"
		   ::);
#endif
#if SPARC_V8
     /* 
      * this should be ok, maybe there is a faster way.
      * Notice that we are using SWAP only for its side effect:
      * atomic operations are ordered with respect to loads and
      * stores.
      */
     int dummy;
     Cilk_swap(&dummy, 0);
#endif
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future writes become visible.
 */
static inline void Cilk_membar_StoreStore(void)
{
     /* don't need to do anything for TSO (Total Store Order) */
}

/*
 *  Ensure that all previous writes are globally visible before any
 *  future reads are performed.
 */
static inline void Cilk_membar_StoreLoad(void)
{
#if SPARC_V9
     asm volatile ("membar #StoreLoad"::);
#endif
#if SPARC_V8
     /* 
      * The only way I know to implement the membar on a V8 sparc
      * is to use swap. Maybe there is a better way... --- athena
      */
     Cilk_fence();
#endif
}

/***********************************************************\
 * Timing functions
\***********************************************************/

#include <sys/time.h>

/* type to hold a high-resolution timer.  Must be an arithmetic type. */
typedef hrtime_t Cilk_time;

/* set after program starts */
DECLARE_PRIVATE_VARIABLE(extern Cilk_time, wall_start, );

static inline void Cilk_init_timers(void)
{
     USE_PRIVATE(wall_start) = gethrtime();
}

static inline Cilk_time Cilk_get_time(void)
{
     return gethrvtime();
}

static inline double Cilk_time_to_sec(Cilk_time t)
{
     return (double) t / 1.0e9;
}

static inline Cilk_time Cilk_get_wall_time(void)
{
     return gethrtime() - USE_PRIVATE(wall_start);
}

static inline double Cilk_wall_time_to_sec(Cilk_time t)
{
     return (double) t / 1.0e9;
}

/***********************************************************\
 * Other random includes
\***********************************************************/

#include <signal.h>
#include <sys/ucontext.h>

/***********************************************************\
 * Typedefs for messaging library
\***********************************************************/

#define MAXMACHINES 100
#define MAX_SEND_QUEUE 2048
#define MAX_MISSING 128

/* header for packets in transit */
typedef struct
{
  void (*handler)();         /* active message function      */
  int size;                  /* # of bytes in message body   */
  int first_seq_no;          /* first seq no of this message */
  int sender;                /* message sender               */
  int seq_no;                /* sequence number from sender  */
  int ack_seq_no;            /* piggy-backed acknowledgement */
} TransitHeader;

/* header for messages waiting to be sent */
typedef struct SendMessage
{
  struct SendMessage *next;
  void (*handler)();         /* active message function      */
  int size;                  /* # of bytes in message body   */
  int buf_size;              /* # of bytes to free when done */
  int seq_no_of_first_chunk; /* sequence # of first chunk    */
  int dest;                  /* destination machine          */
  void *data;                /* data of message (body)       */
} SendMessage;

/* header for packets being reconstructed at receiver */
typedef struct RecvMessage
{
  struct RecvMessage *next;
  int size;                  /* # of bytes in message body   */
  int seq_no_of_first_chunk; /* sequence # of first chunk    */
  int chunks_so_far;         /* # of chunks received         */
  void *data;                /* data of message (body)       */
} RecvMessage;

/* data type for keeping track of which packets need to be sent */
typedef struct
{
  int seq_no;
  SendMessage *msg;
} SendReq;

/* data type for keeping track of which packets we still need from sender */
typedef struct
{
  int sender;                /* message sender               */
  int seq_no;                /* sequence # of missing chunk  */
  Cilk_time gone_missing;    /* time we noticed absence      */
} MissingPacket;

/***********************************************************\
 * Typedefs for cache
\***********************************************************/

/*
 * This number should be the size of the largest virtually indexed
 * cache in the system, divided by its associativity and the
 * system page size.  For example, on SuperSparc processors, there
 * is a 16k virtually indexed direct-mapped L1 cache, and a page
 * size of 8k.  This gives a color value of (16k / 1 / 8k) = 2.
 */
#define VCOLORS 2
