
/***********************************************************\
 * Backing store
\***********************************************************/

#include <shmem-distrib.h>

#define TRACE(x,y)

/***********************************************************\
 * Backing store structures
\***********************************************************/

typedef struct BstorePage
{
  struct BstorePage *next;           /* linked list of pages         */
  int stack_id;                      /* allocating stack id of page  */
  void *page;                        /* virtual address of page      */
  void *data;                        /* page data                    */
} BstorePage;

/*
 * Optimization: we could store diffs at the backing store instead
 * of whole pages.  That way, pages with only a few used locations can
 * be stored cheaply and transferred quickly.  We can merge diffs
 * when writing to the backing store, and send diffs to requesters
 * in raw form (they must be expanded at the destination).
 *    We would probably want some threshold when we switch from a
 * sparse representation like diffs to a dense representation like
 * we have now.
 */

/* hash table of page headers for pages the bstore has allocated */
#define BSTORE_HASH_SIZE 839
DECLARE_MACHINE_VARIABLE(static struct BstorePage **, bstore_page_table, );

DECLARE_MACHINE_VARIABLE(static void *, BSfreelist, );

DECLARE_MACHINE_VARIABLE(static Cilk_mutex, bstore_lock, );

/***********************************************************\
 * Local functions
\***********************************************************/

/*
 * Hash into backing store.
 */
static inline unsigned bstore_hash(int stack_id, void *page)
{
  int pagenum = (size_t)page / USE_MACHINE(pagesize);
  CILK_ASSERT(((unsigned)(pagenum - Machine)) % USE_MACHINE(Machines) == 0);
  return (((unsigned)(pagenum - Machine)) / USE_MACHINE(Machines)) % BSTORE_HASH_SIZE;
}

#if 0

/* increment of pages to allocate for backing store */
/* currently set to 256KB (for 8K pages) */
#define BS_SIZE 32

static void *bstore_new_page()
{
  void *page;
  void *BSpages;
  int i;
  
  page = USE_MACHINE(BSfreelist);
  if (page == NULL)
    {
      /* Allocate some memory for backing store */
      BSpages = (void*) Cilk_malloc_machine_fixed(USE_MACHINE(pagesize) * BS_SIZE);
      if (BSpages == NULL)
	Cilk_die("no bstore memory left!\n");
      
      /* form a linked list of new pages for free list */
      USE_MACHINE(BSfreelist) = BSpages;
      for(i=0; i < BS_SIZE-1; i++)
	*(void**)(BSpages + USE_MACHINE(pagesize) * i) = BSpages + USE_MACHINE(pagesize) * (i+1);
      *(void**)(BSpages + USE_MACHINE(pagesize) * i) = NULL;
      
      page = USE_MACHINE(BSfreelist);
    }
  
  /* splice page out of free list */
  USE_MACHINE(BSfreelist) = *((void**)page);
  
  /* zero page */
  zero_page(page);
  
  return page;
}

static void bstore_free_page(void *page)
{
  CILK_ASSERT(page != NULL);
  
  /* splice into free list */
  *(void**)page = USE_MACHINE(BSfreelist);
  USE_MACHINE(BSfreelist) = page;
}

#endif

static void *bstore_new_page()
{
  void *page;
  page = Cilk_malloc_machine_fixed(USE_MACHINE(pagesize));
  /* zero_page(page); not needed - memory is always zero! KHR */
  return page;
}

/***********************************************************\
 * Interface functions
\***********************************************************/

void Cilk_bstore_init(void)
{
  int i;
  
  USE_MACHINE(bstore_page_table) = Cilk_malloc_machine_fixed(BSTORE_HASH_SIZE * sizeof(BstorePage*));
  
  for(i=0;i<BSTORE_HASH_SIZE;i++)
    USE_MACHINE(bstore_page_table)[i] = NULL;
  USE_MACHINE(BSfreelist) = NULL;
  
  Cilk_mutex_init(&USE_MACHINE(bstore_lock));
}

void Cilk_bstore_done(void)
{
  int i;
  BstorePage *bp, *next;
  
  for(i=0; i<BSTORE_HASH_SIZE; i++)
    for(bp=USE_MACHINE(bstore_page_table)[i]; bp!=NULL; bp=next)
      {
	next = bp->next;
	Cilk_free_machine(bp, sizeof(*bp));
      }
}

void Cilk_bstore_fetch_req(FetchRequest *req)
{
  void *page = req->page;
  int stack_id = req->stack_id;
  BstorePage *bp;
  FetchReply *reply;
  void *data;
  
  Cilk_enter_state(STATE_BSTORE_OP);
  Cilk_mutex_wait(&USE_MACHINE(bstore_lock));
  for(bp = USE_MACHINE(bstore_page_table)[bstore_hash(stack_id, page)]; bp; bp=bp->next)
    if (bp->page == page && bp->stack_id == stack_id)
      {
	data = bp->data;
	CILK_ASSERT(data != NULL);
	Cilk_mutex_signal(&USE_MACHINE(bstore_lock));
	
	reply = Cilk_alloc_msg_buf(sizeof(FetchReply) + USE_MACHINE(pagesize), req->machine);
	copy_page(data, reply->data);
	
	TRACE(TRACE_BSTORE, ("page %p(%d) sent to %d\n", page, stack_id, req->machine));
	
	reply->c = req->c;
	reply->zero = 0;
	reply->timestamp = req->timestamp;
	Cilk_rpc(req->machine, Cilk_cache_fetch_reply, reply, sizeof(*reply) + USE_MACHINE(pagesize));
	Cilk_exit_state(STATE_BSTORE_OP);
	return;
      }
  
  /*
   * If a page isn't present in the table, it must be all zero.
   */
  Cilk_mutex_signal(&USE_MACHINE(bstore_lock));
  TRACE(TRACE_BSTORE, ("zero page %p(%d) sent to %d\n", page, stack_id, req->machine));
  reply = Cilk_alloc_msg_buf(sizeof(FetchReply), req->machine);
  reply->c = req->c;
  reply->zero = 1;
  reply->timestamp = req->timestamp;
  Cilk_rpc(req->machine, Cilk_cache_fetch_reply, reply, sizeof(*reply));
  Cilk_exit_state(STATE_BSTORE_OP);
}

void Cilk_bstore_reconcile_req(ReconcileRequest *req)
{
  void *page = req->page;
  int stack_id = req->stack_id;
  BstorePage *bp;
  BstorePage **t = &USE_MACHINE(bstore_page_table)[bstore_hash(stack_id, page)];
  ReconcileReply *reply;
  void *data;

  Cilk_enter_state(STATE_BSTORE_OP);
  Cilk_mutex_wait(&USE_MACHINE(bstore_lock));
  for(bp = *t; bp; bp=bp->next)
    {
      if (bp->page == page && bp->stack_id == stack_id)
	{
	doit:
	  data = bp->data;
	  CILK_ASSERT(data != NULL);
	  
	  Cilk_mutex_signal(&USE_MACHINE(bstore_lock));
	  
	  apply_diff(req->diff, req->size, data);
	  
	  TRACE(TRACE_BSTORE, ("page %p(%d) reconciled\n", page, stack_id));
	  
	  /* tell reconciler that reconcile is done */
	  reply = Cilk_alloc_msg_buf(sizeof(ReconcileReply), req->machine);
	  reply->c = req->c;
	  Cilk_rpc(req->machine, Cilk_cache_reconcile_reply, reply, sizeof(*reply));
	  
	  Cilk_exit_state(STATE_BSTORE_OP);
	  return;
	}
    }
  
  /* backing store doesn't know about page yet: allocate a new one */
  bp = (BstorePage*)Cilk_malloc_machine(sizeof(BstorePage));
  bp->stack_id = stack_id;
  bp->page = page;
  bp->data = bstore_new_page();
  bp->next = *t;
  *t = bp;
  TRACE(TRACE_BSTORE, ("page %p(%d) added to bstore\n", page, stack_id));
  goto doit;
}

