/* $Id: cilk-internal.h,v 1.2.2.9 1997/12/19 18:20:38 randall Exp $ */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#ifndef _CILK_INTERNAL_H
#define _CILK_INTERNAL_H

/* common definitions used internally by the rts */
#include <limits.h>

/******************************************************
 *  Closures and associated operations
 ******************************************************/
/*
 * If a frame is at the top of its processor stack, these additional
 * informations are recorded.
 */
enum ClosureStatus { CLOSURE_RUNNING = 42, CLOSURE_SUSPENDED,
                     CLOSURE_RETURNING, CLOSURE_READY,
                     CLOSURE_FORWARDING, CLOSURE_NOT_VALID_YET,
                     CLOSURE_STEAL_FAILURE, CLOSURE_DESTROYED};

typedef struct {
  struct Closure *c;     /* local closure pointer */
  int machine;           /* machine id            */
} ClosureID;

/* The order is important here. */
enum AbortStatus { ABORT_ALL = 30 , ALMOST_NO_ABORT, NO_ABORT};

/* 
 * the list of children is not distributed among
 * the children themselves, in order to avoid extra protocols
 * and locking.
 */
/*
 * This structure contains all the data needed in order to execute an
 * inlet.  Some inlets are "complete", other are not.  "Incomplete"
 * inlets are still waiting for data.
 *
 * Each closure has a list of complete inlets, and a list
 * of incomplete inlets.  The list of incomplete inlets
 * is isomorphic to the list of outstanding children,
 * and is also used for that purpose.
 * Complete inlets are ready to go, but are waiting for
 * atomic access to the frame.
 */
struct InletClosure {
     int child_num;                     /* child that structure is for */
     struct InletClosure *next;
  
     int index;
     void (*inlet) (void *, void *, void *);
     int argsize;
     void *inlet_args;
};

struct ReturnedValue {
     int child_num;
     struct ReturnedValue *next;
     
     int size;
     void *value;
     WHEN_CILK_TIMING(Cilk_time cp);
     WHEN_CILK_TIMING(Cilk_time work);
};

typedef struct
{
  WHEN_CILK_TIMING(Cilk_time cp;)
  WHEN_CILK_TIMING(Cilk_time work;)
  int value_size;
  char value[0] __attribute__((aligned(CILK_MAX_ALIGN)));
} ReturnStuff;

typedef struct Closure {
     Cilk_mutex mutex;          /* mutual exclusion lock */
     WHEN_CILK_DEBUG(int mutex_owner;)

     int join_counter; /* number of outstanding children */
     enum ClosureStatus status;
     
     /*
      * return value, valid only 
      * for CLOSURE_RETURNING closures 
      */
     ReturnStuff *return_stuff;

     struct ReturnedValue *returned_values;
     struct InletClosure *incomplete_inlets;  
     /* also, list of outstanding children */

     ClosureID parent; /* the parent closure */
     int child_num;    /* index in list of parent's children        */
     int curr_child;   /* next child number to give out to children */

     StackFrame *frame;       /* rest of the closure */
     enum AbortStatus abort_status; 
     struct ClosureCache *cache; /* see below */

     /* 
      * stuff related to ready deque.  These fields
      * must be managed only by the queue manager in sched.c
      */
     struct Closure *next_ready; 
     struct Closure *prev_ready; 
     WHEN_CILK_DEBUG(int owner_ready_deque;)

     /* miscellanea */
     WHEN_CILK_DEBUG(int magic;)

     ClosureID forward;             /* valid only for CLOSURE_FORWARDING */

     int acquire_level_needed;

     /* critical path and work */
     Cilk_time cp;      
     Cilk_time work;

     CILK_CACHE_LINE_PAD;
} Closure;

/******************************************************
 *  Options processing
 ******************************************************/

typedef struct
{
  int nmach;		/* number of machines				*/
  int stackdepth;	/* depth of scheduling stack			*/
  size_t heap;		/* size of user heap				*/
  size_t cache;		/* size of shared-memory cache			*/
  char cachedir[PATH_MAX+1]; /* directory to allocate cache file        */
  size_t pagesize;	/* size of system page				*/
  size_t rtsheap;	/* size of internal rts heap (per machine)	*/
  size_t bstore;	/* size of backing store (total)                */
  int statlevel;	/* statistics gathering level			*/
  int yieldslice;	/* lower priority when stealing			*/
  int dump_core;
  int argc;
  char **argv;          /* options for user program                     */

  /*
   * The following options are not settable by the user - they
   * are computed by the system and passed to each machine.
   */
  int start_proc_no;    /* process # of first process on this machine	*/
  int permach;		/* number of processes on this machine		*/
  void *heap_ptr;       /* heap pointer                                 */
  int nproc;            /* total # of processes on all machines         */
} Options;

/*
 * Machine name and size for each machine on command line.  Filled in
 * by Cilk_parse_command_line.
 */
extern char *Cilk_option_machine[100];
extern int Cilk_option_nproc[100];

#define DEFAULT_OPTIONS \
(Options) \
{ \
    0, \
    CILK_DEFAULT_STACK_DEPTH, \
    CILK_DEFAULT_HEAPSIZE, \
    CILK_DEFAULT_CACHESIZE, \
    "/tmp", \
    0, \
    CILK_DEFAULT_RTSHEAPSIZE, \
    CILK_DEFAULT_BSTORESIZE, \
    0, \
    1, \
    0, \
    0, \
    NULL, \
}

/* default number of processes per machine (including comm. processes) */
#define DEFAULT_NPROC 2

/* command-line parser */
extern int Cilk_parse_command_line(Options *options, int argc, char *argv[]);

/* global malloc */
extern void Cilk_malloc_init_machine(const Options *options);
extern void Cilk_malloc_init_local(const Options *options);

/* per-machine malloc */
extern void *Cilk_malloc_machine(size_t size);
extern void Cilk_free_machine(void *p, size_t size);
extern void Cilk_pretend_to_free_machine(void *p, size_t size);
extern void *Cilk_malloc_machine_fixed(size_t size);
extern void Cilk_malloc_machine_init(const Options *options);
extern void Cilk_malloc_machine_done(const Options *options);
extern void Cilk_malloc_done_local(const Options *options);

/* scheduler */
extern void Cilk_scheduler_init_machine(const Options *options);
extern void Cilk_scheduler_init_local(const Options *options);
extern Closure *Cilk_create_initial_thread(int argc, char *argv[]);
extern void Cilk_scheduler(Closure *t);

/* other variables */
DECLARE_PRIVATE_VARIABLE(extern int, yield_slice, );
DECLARE_PRIVATE_VARIABLE(extern int, trace_mode, );
DECLARE_MACHINE_VARIABLE(extern int, done, );
DECLARE_MACHINE_VARIABLE(extern int, retval, );

DECLARE_PRIVATE_VARIABLE(extern void *, infofile, );


/* shared memory */
DECLARE_MACHINE_VARIABLE(extern void *, heap, );
DECLARE_MACHINE_VARIABLE(extern size_t, heap_size, );
DECLARE_PRIVATE_VARIABLE(extern void *, shared_region_begin, );
DECLARE_PRIVATE_VARIABLE(extern void *, shared_region_end, );

extern void Cilk_memory_init_global(Options *options);
extern void Cilk_memory_init_machine(const Options *options);
extern void Cilk_memory_init_local(const Options *options);
extern void Cilk_memory_done_machine(const Options *options);
extern void Cilk_memory_done_global(const Options *options);
extern void Cilk_memory_release(void *addr, size_t len, int level);
extern void Cilk_memory_release_from(void *addr, size_t len, int from, int level);
extern void Cilk_memory_acquire(void *addr, size_t len, int level);
extern void Cilk_memory_acquire_from(void *addr, size_t len, int from, int level);

/******************************************************
 *  Architecture-specific interface
 ******************************************************/

/*
 * Functions defined in the architecture-specific file and used
 * only by the runtime system.
 */
extern void Cilk_map_shared_segment(void);
extern void Cilk_map_private_segment(void);
extern void Cilk_map_machine_segment(void);

extern void *Cilk_map_shared(size_t len);
extern int Cilk_arch_specific_options(int i, int argc, char *argv[]);
extern void Cilk_arch_specific_init_global(Options *options);
extern void Cilk_arch_specific_init_machine(const Options *options);
extern void Cilk_arch_specific_init_local(const Options *options);
extern void Cilk_arch_specific_done_global(const Options *options);
extern void Cilk_arch_specific_done_machine(const Options *options);
extern void Cilk_arch_specific_done_local(const Options *options);
extern void Cilk_create_children(const Options *options, void (*child)(const Options *, int));
extern void Cilk_create_machines(Options *options);
extern void Cilk_machine_main(const Options *options, int id);
extern void Cilk_start_machine(int argc, char *argv[]);
extern void Cilk_lower_priority(void);
extern void Cilk_raise_priority(void);

/* architecture-specific memory mapping support */
void Cilk_setup_fault_handler(void);
void *Cilk_map_alloc(void *addr, size_t size);
void *Cilk_setup_cache(size_t size, const Options *options);
#define CILK_MAP_FAULTING 0
#define CILK_MAP_READONLY 1
#define CILK_MAP_READWRITE 2
#define CILK_MAP_NOSTAT 16
void Cilk_map_cache_page(void *addr, void *cache_addr, int mode);
void Cilk_unmap_region(void *addr, size_t size);
void Cilk_send_eviction_notice(int i);

/* callbacks from arch-specific files to cilk files */
int Cilk_fault_handler(void *addr, int write);
void Cilk_cache_evict_pages(void);

/* message passing */
extern void Cilk_message_passing_init_global(Options *options);
extern void Cilk_message_passing_init_machine(const Options *options);
extern void Cilk_message_passing_init_machine_globalproc(Options *options, int *mconn);
extern void Cilk_message_passing_init_local(const Options *options);
extern void *Cilk_alloc_msg_buf(int size, int dest_pn);
extern void Cilk_free_msg_buf(void *data, int size, int dest_pn);
extern void Cilk_rpc(int dest_pn, void (*handler)(), void *data, int size);
extern void Cilk_rpc_with_size(int dest_pn, void (*handler)(), void *data, int buf_size, int size);
void Cilk_poll(void);

/* barriers */
extern void Cilk_barrier(void);
extern void Cilk_barrier_init_machine(const Options *options);
extern void Cilk_barrier_init_local(const Options *options);


/******************************************************
 *  Other stuff
 ******************************************************/

/* statistics */
#if CILK_STATS

enum { 
     EVENT_THREAD,
     EVENT_STEAL_ATTEMPT,
     EVENT_STEAL,
     EVENT_STEAL_NO_DEKKER,
     EVENT_STEAL_EMPTY_DEQUE,
     EVENT_STEAL_RETURNING,
     EVENT_STEAL_ABORT,
     EVENT_STEAL_REMOTE_SUCCESS,
     EVENT_STEAL_REMOTE_FAILURE,

     EVENT_ABORT_SLOW,
     EVENT_ABORT_STANDALONE,
     EVENT_ABORT_READY_SUSPEND,
     EVENT_ABORT_READY_RETURN,

     EVENT_SUSPEND,
     EVENT_CILK_SYNC,

     EVENT_EXCEPTION,
     EVENT_EXCEPTION_STEAL,
     EVENT_EXCEPTION_ABORT,
     EVENT_EXCEPTION_ABORT_RETURN,
     EVENT_EXCEPTION_ABORT_SUSPEND,
     EVENT_EXCEPTION_OTHER,

     EVENT_RETURN_SLOW,
     EVENT_PROVABLY_GOOD_STEAL,
     EVENT_RETURN_ENQUEUE,
     EVENT_POLL_INLETS,
     EVENT_CLOSURE_CREATE,

     EVENT_FETCH,
     EVENT_FETCH_ZERO,
     EVENT_FETCH_NULL,
     EVENT_RECONCILE,
     EVENT_RECONCILE_NULL,

     EVENT_USER0,
     EVENT_USER1,
     EVENT_USER2,
     EVENT_USER3,
     EVENT_USER4,
     EVENT_USER5,
     EVENT_USER6,
     EVENT_USER7,
     EVENT_NTYPES,  		/* this must be last */
};

extern void Cilk_event(int type);
extern void Cilk_stats_init_local(const Options *options);
extern void Cilk_stats_init_machine(const Options *options);
extern void Cilk_stats_init_global(Options *options);
extern void Cilk_stats_done_local(const Options *options);
extern void Cilk_stats_done_machine(const Options *options);
extern void Cilk_stats_done_global(const Options *options);

#else

#define Cilk_event(a)

#endif

/* timing */
#if CILK_TIMING
extern void Cilk_enter_state(int state);
extern void Cilk_exit_state(int state);
extern double Cilk_compute_work(int active_size);

extern void Cilk_timing_init_local(const Options *options);
extern void Cilk_timing_init_machine(const Options *options);
extern void Cilk_timing_init_global(Options *options);
extern void Cilk_timing_done_local(const Options *options);
extern void Cilk_timing_done_machine(const Options *options);
extern void Cilk_timing_done_global(const Options *options);


enum {
     STATE_TOTAL,
     STATE_WORKING,
     STATE_WAITING_FOR_LOCK,
     STATE_STEALING,
     STATE_RETURNING,
     STATE_ABORT_RECURSIVE,
     STATE_COMPUTE_DIFF,
     STATE_APPLY_DIFF,
     STATE_COPY_PAGE,
     STATE_TWIN_PAGE,
     STATE_ZERO_PAGE,
     STATE_WAITING_FOR_FETCH,
     STATE_WAITING_FOR_RECONCILE,
     STATE_BSTORE_OP,
     STATE_USER0,                /* for the user */
     STATE_USER1,
     STATE_USER2,
     STATE_USER3,
     STATE_USER4,
     STATE_USER5,
     STATE_USER6,
     STATE_USER7,
     TIME_NSTATES		/* this must be last */
};

#else

#define Cilk_enter_state(state)
#define Cilk_exit_state(state)
#define Cilk_compute_work() (0.0)

#endif				/* timing */

/*
 * functions used by the scheduler and by other protocols
 * operating on the Closures tree
 */
extern Closure *Cilk_Closure_create(void);

#endif				/* _CILK_INTERNAL_H */
