/*
 * $Id: machine-malloc.c,v 1.1.2.3 1997/12/10 02:10:04 randall Exp $
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>

typedef struct Bin
{
  int maxsize;
  void *freelist;
} Bin;

#define NBINS 28

DECLARE_MACHINE_VARIABLE(static struct Bin *, bins, );

DECLARE_MACHINE_VARIABLE(static Cilk_mutex, lock, );

DECLARE_MACHINE_VARIABLE(static void *, arena, );
DECLARE_MACHINE_VARIABLE(static void *, end_arena, );

DECLARE_MACHINE_VARIABLE(static int, used, );
DECLARE_MACHINE_VARIABLE(static int, max, );

static Bin *find_bin(size_t size)
{
  Bin *bin;
  
  for(bin = USE_MACHINE(bins); bin < USE_MACHINE(bins) + NBINS; bin++)
    {
      if (size <= bin->maxsize)
	return bin;
    }
  
  Cilk_die("size too big for any bin!");
  return NULL;
}

void *Cilk_malloc_machine(size_t size)
{
  Bin *bin;
  void *p;
  
  bin = find_bin(size);
  
  Cilk_mutex_wait(&USE_MACHINE(lock));
  
  WHEN_CILK_DEBUG(USE_MACHINE(used) += size;
		  if (USE_MACHINE(used) > USE_MACHINE(max))
		    USE_MACHINE(max) = USE_MACHINE(used));
  
  p = bin->freelist;
  if (p != NULL)
    {
      bin->freelist = *((void**)p);
    }
  else
    {
      p = USE_MACHINE(arena);
      USE_MACHINE(arena) = p + bin->maxsize;
      CILK_CHECK(USE_MACHINE(arena) <= USE_MACHINE(end_arena),
		 ("Cilk machine malloc() memory exhausted, can't alloc %d bytes.  Ask for more memory with `-bstore'.\n",
		  size));
    }
  
  Cilk_mutex_signal(&USE_MACHINE(lock));
  
  return p;
}

void *Cilk_malloc_machine_fixed(size_t size)
{
  void *p;
  
  size = ALIGN(size, CILK_MAX_ALIGN);
  
  Cilk_mutex_wait(&USE_MACHINE(lock));
  
  p = USE_MACHINE(arena);
  USE_MACHINE(arena) = p + size;
  CILK_CHECK(USE_MACHINE(arena) <= USE_MACHINE(end_arena),
	     ("Cilk machine malloc() memory exhausted, can't alloc %d bytes.  Ask for more memory with `-bstore'.\n",
	      size));
  
  Cilk_mutex_signal(&USE_MACHINE(lock));
  
  return p;
}

void Cilk_free_machine(void *p, size_t size)
{
  Bin *bin = find_bin(size);

  Cilk_mutex_wait(&USE_MACHINE(lock));

  WHEN_CILK_DEBUG(USE_MACHINE(used) -= size);
  
  *((void**)p) = bin->freelist;
  bin->freelist = p;
  
  Cilk_mutex_signal(&USE_MACHINE(lock));
  
  /*
   * Note: we could use compare and swap directly here, and in malloc.
   */
}

void Cilk_pretend_to_free_machine(void *p, size_t size)
{
  Cilk_mutex_wait(&USE_MACHINE(lock));
  WHEN_CILK_DEBUG(USE_MACHINE(used) -= size);
  Cilk_mutex_signal(&USE_MACHINE(lock));
}

void Cilk_malloc_machine_init(const Options *options)
{
  int i;
  int size;
  
  size = ALIGN(options->rtsheap + options->bstore/options->nmach, options->pagesize);
  
  USE_MACHINE(arena) = Cilk_map_shared(size);
  CILK_ASSERT(USE_MACHINE(arena) == ALIGN(USE_MACHINE(arena), CILK_MAX_ALIGN));
  USE_MACHINE(end_arena) = USE_MACHINE(arena) + size;
  
  Cilk_mutex_init(&USE_MACHINE(lock));
  
  USE_MACHINE(bins) = Cilk_malloc_machine_fixed(NBINS * sizeof(Bin));
  for(i=0; i<NBINS; i++)
    {
      USE_MACHINE(bins)[i].maxsize = (1<<i)*CILK_MAX_ALIGN;
      USE_MACHINE(bins)[i].freelist = NULL;
    }
}


void Cilk_malloc_machine_done(const Options *options)
{
  WHEN_CILK_DEBUG(Cilk_dprintf_machine("used:%d max:%d\n", USE_MACHINE(used), USE_MACHINE(max)));
  /* CILK_ASSERT_MACHINE(USE_MACHINE(used) == 0); */
}
