/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk.h>
#include <cilk-internal.h>
#include <cilk-lib.h>

#include <sys/imc.h>

#define CILK_MEMCHAN_KEY_BASE 1295666

#define BUFSIZE (256*1024)

/* This code can sustain double the message bandwidth (~66 MB/s
 * instead of ~33 MB/s) if "loopback" is turned off.  If loopback is
 * off, however, you cannot run two virtual machines on one physical
 * machine.  Loopback is currently disabled by default.
 *
 * Could we auto-detect this situation at some point???
 */
#if 1
#undef IMC_LOOPBACK
#define IMC_LOOPBACK NULL
#endif

#define Machines USE_MACHINE(Machines)
#define Machine USE_MACHINE(Machine)

struct returnbuf
{
  int off;
  int size;
};

struct freebuf
{
  struct freebuf *next;
  int off;
  int size;
};

struct links
{
  Cilk_mutex poll_lock;
  int master_poll;
  struct link
  {
    // set by a message sender
    void (*handler)();       // handler to run
    int msg_off;             // offset of message data in buffer
    int msg_size;            // size of message data
    
    // can free two message buffers with each message (off==-1 if none to free)
    int return1_off;
    int return1_size;
    int return2_off;
    int return2_size;
    
    // set to acknowledge message or free buffer space
    int outstanding;         // message outstanding
    
    /* doesn't really need to be here, but convenient to keep with above */
    Cilk_mutex send_lock;    // lock to obtain before sending
    int returnhead, returntail;
    struct returnbuf returnlist[64];  // list of buffers not returned yet
    Cilk_mutex free_lock;    // lock to obtain to alloc/free buffer space
    struct freebuf *free_list;  // list of free buffers for sending

    void *in_buf;
    void *out_buf;
    
    // other randomness
    int init;                // initialization flag
  } link[MAXMACHINES];
};

DECLARE_MACHINE_VARIABLE(static Cilk_mutex, init_lock, );
DECLARE_MACHINE_VARIABLE(static volatile int, init_count, );

/* local link space mapped from each machine */
DECLARE_MACHINE_VARIABLE(static volatile struct links *, in_links, );

/* remote link space mapped to each machine */
DECLARE_MACHINE_VARIABLE(static volatile struct links *, out_links, [MAXMACHINES]);

/* sender buffer management */
static int alloc_buf(volatile struct link *link, int size)
{
  int off;
  int remain;
  struct freebuf *f, * volatile *g;
  
  /* search buffers for a big enough free one */
  for(g = &link->free_list, f = *g; f; g = &f->next, f = *g)
    {
      remain = f->size - size;
      if (remain >= 0)
	{
	  off = f->off;
	  if (remain == 0)  /* exact size */
	    {
	      *g = f->next;
	      Cilk_free_machine(f, sizeof(*f));
	    }
	  else
	    {
	      f->off = off + size;
	      f->size = remain;
	    }
	  return off;
	}
    }
  Cilk_die("couldn't find any free memory!\n");
  return 0;  /* keep gcc happy */
}

static void free_buf(volatile struct link *link, int off, int size)
{
  struct freebuf *f, * volatile *g;
  
  /* try to find adjoining range in free list */
again:
  for(g = &link->free_list, f = *g; f; g = &f->next, f = *g)
    {
      /* if we find one, get combined region and try again */
      if (f->off == off + size)  /* adjoins beginning of f */
	{
	  /* off unchanged */
	  size = size + f->size;
	  *g = f->next;
	  Cilk_free_machine(f, sizeof(*f));
	  goto again;
	}
      else if (off == f->off + f->size)  /* adjoins end of f */
	{
	  off = f->off;
	  size = f->size + size;
	  *g = f->next;
	  Cilk_free_machine(f, sizeof(*f));
	  goto again;
	}
    }
  /* no adjoining region: add to list */
  f = Cilk_malloc_machine(sizeof(*f));
  f->off = off;
  f->size = size;
  f->next = link->free_list;
  link->free_list = f;
}

static void run_handler(void (*handler)(), void *data)
{
#if CILK_TIMING
  void Cilk_uninclude_time(Cilk_time process, Cilk_time wall);
  Cilk_time begin = Cilk_get_time();
  Cilk_time begin_wall = Cilk_get_wall_time();
#endif
  
  handler(data);
  
#if CILK_TIMING
  Cilk_uninclude_time(Cilk_get_time() - begin,
		      Cilk_get_wall_time() - begin_wall);
#endif
}

void Cilk_rpc(int dest_pn, void (*handler)(), void *data, int buf_size)
{
  volatile struct link *in;
  volatile struct link *out;
  int h,t;
  
  /* shortcut loopback */
  if (dest_pn == Machine)
    {
      run_handler(handler, data);
      Cilk_free_msg_buf(data, buf_size, dest_pn);
      return;
    }
  
  buf_size = ALIGN(buf_size, CILK_MAX_ALIGN);
  
  in = &USE_MACHINE(in_links)->link[dest_pn];
  out = &USE_MACHINE(out_links)[dest_pn]->link[Machine];
  
again:
  /* wait until previous message is acknowledged */
  while(in->outstanding)
    ;/* Cilk_poll();           // do we need to poll here??? */
  
  Cilk_mutex_wait(&in->send_lock);
  
  if(in->outstanding)
    {
      Cilk_mutex_signal(&in->send_lock);
      goto again;
    }
  
  /* set outstanding flag in local link */
  in->outstanding = 1;
  
  /* set pointer in remote link to message */
  out->msg_off = data - in->out_buf;
  out->msg_size = buf_size;

  /* return some message buffers */
  h = in->returnhead;
  t = in->returntail;
  if (h != t)
    {
      out->return1_off = in->returnlist[t].off;
      out->return1_size = in->returnlist[t].size;
      t = (t+1) & 63;
      if (h != t)
	{
	  out->return2_off = in->returnlist[t].off;
	  out->return2_size = in->returnlist[t].size;
	  t = (t+1) & 63;
	}
      else
	out->return2_off = -1;
      in->returntail = t;
    }
  else
    {
      out->return1_off = -1;
      out->return2_off = -1;
    }
  Cilk_membar_StoreStore();
  out->handler = handler;   // handler write must be last!
  
  /* set master flag to tell destination that a message is available */
  Cilk_membar_StoreStore();
  USE_MACHINE(out_links)[dest_pn]->master_poll = 1;
  
  Cilk_mutex_signal(&in->send_lock);
}

void Cilk_rpc_with_size(int dest_pn, void (*handler)(), void *data, int buf_size, int size)
{
  Cilk_rpc(dest_pn, handler, data, buf_size);
}

void Cilk_poll(void)
{
  int i;
  volatile struct links *links = USE_MACHINE(in_links);
  
  /* check to see if we have any messages */
  if (!links->master_poll)
    return;
  
  Cilk_mutex_wait(&links->poll_lock);
  
  if (!links->master_poll)
    {
      Cilk_mutex_signal(&links->poll_lock);
      return;
    }
  
  /* There's may be a message somewhere - look for it */
  for(i=0; i<Machines; i++)
    {
      volatile struct link *in = &links->link[i];
      void (*handler)() = in->handler;
      if (handler)
	{
	  /* Found a message.  Get it and run it */
	  int off = in->msg_off;
	  int size = in->msg_size;
	  int return1_off = in->return1_off;
	  int return1_size = in->return1_size;
	  int return2_off = in->return2_off;
	  int return2_size = in->return2_size;
	  volatile struct link *ack;
	  int h;
	  
	  /* reset to allow another poller */
	  in->handler = 0;
	  Cilk_mutex_signal(&links->poll_lock);
	  
	  /* acknowledge to allow another incoming message from i */
	  ack = &USE_MACHINE(out_links)[i]->link[Machine];
	  ack->outstanding = 0;
	  
	  /* run handler */
	  run_handler(handler, in->in_buf + off);
	  
	  /* add just-used buffer to list of returned buffers */
	  Cilk_mutex_wait(&in->send_lock);
	  h = in->returnhead;
	  in->returnlist[h].off = off;
	  in->returnlist[h].size = size;
	  h = (h+1) & 63;
	  if (h == in->returntail)
	    Cilk_die("filled return buffer queue!\n");
	  in->returnhead = h;
	  Cilk_mutex_signal(&in->send_lock);
	  
	  /* free returned buffers */
	  if (return1_off >= 0)
	    {
	      Cilk_mutex_wait(&in->free_lock);
	      free_buf(in, return1_off, return1_size);
	      if (return2_off >= 0)
		free_buf(in, return2_off, return2_size);
	      Cilk_mutex_signal(&in->free_lock);
	    }
	  
	  return;
	}
    }
  
  /* if we make it all the way through, we can reset the master_poll flag,
   * and then check once more.  If there is another message, just set
   * master flag to 1 again and quit.
   */
  links->master_poll = 0;
  Cilk_membar_StoreLoad();
  for(i=0; i<Machines; i++)
    if (links->link[i].handler)
      {
	/* Note: we could jump to the top of this function again.  It
	 * is probably not a good idea, however, because although
	 * there may be more messages, the message this particular
	 * poller was waiting for may have arrived.  It's probably
	 * better to return to the caller and let him check, and call
	 * poll again if necessary.
	 */
	links->master_poll = 1;
	break;
      }
  
  Cilk_mutex_signal(&links->poll_lock);
}

void *Cilk_alloc_msg_buf(int size, int dest_pn)
{
  void *buf;
  
  /* Cilk_dprintf("allocd buf %d %d\n", size, dest_pn); */
  if (dest_pn == Machine)
    {
      buf = Cilk_malloc_machine(size);
    }
  else
    {
      volatile struct link *link = &USE_MACHINE(in_links)->link[dest_pn];
      int off;
      
      size = ALIGN(size, CILK_MAX_ALIGN);
      
      Cilk_mutex_wait(&link->free_lock);
      off = alloc_buf(link, size);
      Cilk_mutex_signal(&link->free_lock);
      
      buf = link->out_buf + off;
    }
  return buf;
}

void Cilk_free_msg_buf(void *data, int size, int dest_pn)
{
  if (dest_pn == Machine)
    {
      Cilk_free_machine(data, size);
    }
  else
    {
      volatile struct link *link = &USE_MACHINE(in_links)->link[dest_pn];
      Cilk_mutex_wait(&link->free_lock);
      free_buf(link, data - link->out_buf, size);
      Cilk_mutex_signal(&link->free_lock);
    }
}

void Cilk_message_passing_init_global(Options *options)
{
}

void Cilk_message_passing_init_machine(const Options *options)
{
  /* Memory channel init should really go here, but for some reason it
   * doesn't last across process forks.  Thus, it is done in
   * Cilk_message_passsing_init_local instead.
   */
}

void Cilk_message_passing_init_machine_globalproc(Options *options, int *mconn)
{
}

void Cilk_message_passing_init_local(const Options *options)
{
  int i;
  int res;
  imc_asid_t link_rgn[MAXMACHINES];
  imc_asid_t outbuf_rgn[MAXMACHINES];
  imc_asid_t inbuf_rgn[MAXMACHINES];
  int key = CILK_MEMCHAN_KEY_BASE;
  volatile struct links *links;
  
  /* allocate link regions */
  for(i=0; i<Machines; i++)
    {
      res = imc_asalloc(key + i, sizeof(struct links), IMC_URW, NULL, &link_rgn[i], 0);
      if (res != IMC_SUCCESS)
	Cilk_die("bad imc_asalloc %d %d\n", i, res);
    }
  
  /* attach to inbound link region */
  res = imc_asattach(link_rgn[Machine], IMC_RECEIVE, IMC_SHARED, IMC_LOOPBACK, (caddr_t*)&links);
  if (res != IMC_SUCCESS)
    Cilk_die("can't attach inbound links %d\n", res);
  USE_MACHINE(in_links) = links;
  
//Cilk_dprintf("attached in link to %p\n", links);

  /* initialize inbound links */
  {
    Cilk_mutex_init(&links->poll_lock);
    links->master_poll = 0;
    for(i=0; i<Machines; i++)
      {
	volatile struct link *link = &links->link[i];
	link->handler = NULL;
	link->msg_off = 0;
	link->msg_size = 0;
	link->outstanding = 0;
	Cilk_mutex_init(&link->send_lock);
	link->returnhead = 0;
	link->returntail = 0;
	Cilk_mutex_init(&link->free_lock);
	link->free_list = Cilk_malloc_machine(sizeof(*link->free_list));
	link->free_list->next = 0;
	link->free_list->off = 0;
	link->free_list->size = BUFSIZE;
	link->init = 0;
      }
  }
  
  /* attach to outbound link regions */
  for(i=0; i<Machines; i++)
    if (i == Machine)
      USE_MACHINE(out_links)[i] = NULL;
    else
      {
	res = imc_asattach(link_rgn[i], IMC_TRANSMIT, IMC_SHARED, IMC_LOOPBACK, (caddr_t*)&USE_MACHINE(out_links)[i]);
	if (res != IMC_SUCCESS)
	  Cilk_die("can't attach outbound links %d %d\n", i, res);
//Cilk_dprintf("attached out link %d to %p\n", i, USE_MACHINE(out_links)[i]);
      }
  
  
  key += Machines;
  
  
  /* allocate buffer regions (P out/in for each machine).  Buffers
   * are numbered in source-major order. */
  
  /* allocate out buffers */
  for(i=0; i<Machines; i++)
    if (i == Machine)
      outbuf_rgn[i] = 0;
    else
      {
	res = imc_asalloc(key + Machine * Machines + i, BUFSIZE, IMC_URW, NULL, &outbuf_rgn[i], 0);
	if (res != IMC_SUCCESS)
	  Cilk_die("bad outbuffer imc_asalloc %d %d\n", i, res);
      }
  
  /* allocate in buffers */
  for(i=0; i<Machines; i++)
    if (i == Machine)
      inbuf_rgn[i] = 0;
    else
      {
	res = imc_asalloc(key + i * Machines + Machine, BUFSIZE, IMC_URW, NULL, &inbuf_rgn[i], 0);
	if (res != IMC_SUCCESS)
	  Cilk_die("bad inbuffer imc_asalloc %d %d\n", i, res);
      }
  
  /* attach to out buffers */
  for(i=0; i<Machines; i++)
    if (i == Machine)
      links->link[i].out_buf = NULL;
    else
      {
	res = imc_asattach(outbuf_rgn[i], IMC_TRANSMIT, IMC_SHARED, IMC_LOOPBACK, (caddr_t*)&links->link[i].out_buf);
	if (res != IMC_SUCCESS)
	  Cilk_die("can't attach outbound buf %d %d\n", i, res);
      }
  
  /* attach to in buffers */
  for(i=0; i<Machines; i++)
    if (i == Machine)
      links->link[i].in_buf = NULL;
    else
      {
	res = imc_asattach(inbuf_rgn[i], IMC_RECEIVE, IMC_SHARED, IMC_LOOPBACK, (caddr_t*)&links->link[i].in_buf);
	if (res != IMC_SUCCESS)
	  Cilk_die("can't attach inbound buf %d %d\n", i, res);
      }
  
  
  key += Machines * Machines;
  
  
  /* barrier to make sure everyone has initialized links before
   * anyone starts sending messages.
   */
#define MAGIC 0x778899aa
  
  Cilk_mutex_wait(&USE_MACHINE(init_lock));
  USE_MACHINE(init_count)++;
  if (USE_MACHINE(init_count) == USE_MACHINE(Processes))
    {
    loop:
      /* write magic number to each machine */
      for(i=0; i<Machines; i++)
	if (i != Machine)
	  USE_MACHINE(out_links)[i]->link[Machine].init = MAGIC;
      
      /* check for magic number from each machine */
      for(i=0; i<Machines; i++)
	if (i != Machine)
	  if (links->link[i].init != MAGIC)
	    goto loop;
      
      /* write one more time */
      for(i=0; i<Machines; i++)
	if (i != Machine)
	  USE_MACHINE(out_links)[i]->link[Machine].init = MAGIC;
      
      /* signal rest of processes on machine that initialization is done */
      USE_MACHINE(init_count) = 0;
    }
  Cilk_mutex_signal(&USE_MACHINE(init_lock));
  
  /* wait for rest of processes on machine to initialize */
  while(USE_MACHINE(init_count) != 0);
}
