#include <shmem-distrib.h>

/***********************************************************\
 * Shared-memory initialization
\***********************************************************/

DECLARE_PRIVATE_VARIABLE(void *, shared_region_begin, );
DECLARE_PRIVATE_VARIABLE(void *, shared_region_end, );

void Cilk_memory_init_global(Options *options)
{
  /* allocate range of addresses to fault on */
  options->heap = ALIGN(options->heap, options->pagesize);
  options->heap_ptr = Cilk_map_alloc(NULL, options->heap);
}

void Cilk_memory_init_machine(const Options *options)
{
  USE_MACHINE(pagesize) = options->pagesize;
  USE_MACHINE(heap) = options->heap_ptr;
  USE_MACHINE(heap_size) = options->heap;
  
  Cilk_map_alloc(options->heap_ptr, options->heap);
  
  Cilk_malloc_machine_init(options);
  
  Cilk_cache_init(options);
  Cilk_bstore_init();
}

void Cilk_memory_init_local(const Options *options)
{
  /* install fault handler */
  Cilk_setup_fault_handler();
  
  /* map user shared segment correctly */
#if CILK_MANAGE_SHARED_SEGMENT
  {
    size_t pagesize = Cilk_get_page_size();
    void *startmem, *endmem;
    
    startmem = &Cilk_all_vars.shared;
    endmem = startmem + sizeof(Cilk_all_vars.shared);
    
    /* round start and end outward to page boundaries */
    startmem = ALIGN(startmem - pagesize + 1, pagesize);
    endmem = ALIGN(endmem, pagesize);
    
    USE_PRIVATE(shared_region_begin) = startmem;
    USE_PRIVATE(shared_region_end) = endmem;
  }
#else
  {
    extern _end_Cilk_shared;	/* linker_defined symbols */
    extern _begin_Cilk_shared;
    USE_PRIVATE(shared_region_begin) = &_begin_Cilk_shared;
    USE_PRIVATE(shared_region_end) = &_end_Cilk_shared;
  }
#endif
  CILK_ASSERT((size_t)USE_PRIVATE(shared_region_begin) % USE_MACHINE(pagesize) == 0);
  CILK_ASSERT((size_t)USE_PRIVATE(shared_region_end) % USE_MACHINE(pagesize) == 0);
  
  if (Self != 0)
    Cilk_unmap_region(USE_PRIVATE(shared_region_begin), USE_PRIVATE(shared_region_end) - USE_PRIVATE(shared_region_begin));
  
  /* the rest is done in invoke_main.c, for Self == 0:
   * 1. Copy shared region to somewhere else.
   * 2. Unmap shared region.
   * 3. Copy shared region back.
   */
}

void Cilk_memory_done_machine(const Options *options)
{
  Cilk_bstore_done();
}

void Cilk_memory_done_global(const Options *options)
{
}



/*
 *                   ACQUIRE
 *  Make sure we get an up-to-date version of all
 * memory locations <addr> through <addr>+<len>-1.
 */
/*
 *                   RELEASE
 * Make sure that future acquirers of the memory
 * locations <addr> through <addr>+<len>-1 get a version
 * of these locations that is at least as up-to-date as
 * the version being released.
 */

void Cilk_memory_release(void *addr, size_t len, int depth)
{
  CILK_ASSERT(depth >= 1 && depth <= 2);
  if (depth >= 1)
    {
      Cilk_fence();
    }
  if (depth >= 2)
    {
      Cilk_reconcile_cache();
    }
}

void Cilk_memory_release_from(void *addr, size_t len, int from, int depth)
{
  CILK_ASSERT(depth == 2);
  Cilk_reconcile_cache_from(from);
}

void Cilk_memory_acquire(void *addr, size_t len, int depth)
{
  CILK_ASSERT(depth >= 1 && depth <= 2);
  if (depth >= 2)
    {
      Cilk_flush_cache();
    }
  if (depth >= 1)
    {
      Cilk_fence();
      Cilk_cache_acquire_local();
    }
}

/* acquire from process <from> on local machine */
void Cilk_memory_acquire_from(void *addr, size_t len, int from, int depth)
{
  CILK_ASSERT(depth == 1);
  Cilk_fence();
  Cilk_cache_acquire_local_from(from);
}
