#ifndef _SHMEM_DISTRIB_H
#define _SHMEM_DISTRIB_H

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <stdio.h>
#include <strings.h>

#define Machine USE_MACHINE(Machine)
#define Process USE_PRIVATE(Process)

/*
 * Architecture-specific variables
 */
DECLARE_MACHINE_VARIABLE(extern int, pagesize, );

/* internal stuff */
typedef struct
{
  int offset;
  int val;
} RecEntry;

typedef struct
{
  void *page;              /* page id                  */
  int stack_id;            /* cactus stack id          */
  int machine;             /* requesting machine       */
  int timestamp;           /* timestamp of request     */
  struct CachePage *c;     /* cache page identifier    */
} FetchRequest;

typedef struct
{
  struct CachePage *c;     /* cache page identifier    */
  int zero;                /* set if page is all zero  */
  int timestamp;           /* timestamp of request     */
  int data[0];             /* data for page (if !zero) */
} FetchReply;


typedef struct
{
  void *page;              /* page id                  */
  int stack_id;            /* cactus stack id          */
  int machine;             /* requesting machine       */
  int size;                /* # of reconcile entries   */
  struct CachePage *c;     /* cache page identifier    */
  RecEntry diff[0];
} ReconcileRequest;

typedef struct
{
  struct CachePage *c;     /* cache page identifier    */
} ReconcileReply;


/***********************************************************\
 * Utilities
\***********************************************************/

static inline void zero_page(void *page)
{
  Cilk_enter_state(STATE_ZERO_PAGE);
  
#if 0
  {
    int i = USE_MACHINE(pagesize) - sizeof(int);
    while(i >= 0)
      {
	*(int*)(page + i) = 0;
	i -= sizeof(int);
      }
  }
#else
  bzero(page, USE_MACHINE(pagesize));
#endif
  
  Cilk_exit_state(STATE_ZERO_PAGE);
}

static inline void copy_page(void *src, void *dest)
{
  Cilk_enter_state(STATE_COPY_PAGE);
  
#if 0
  {
    int i = USE_MACHINE(pagesize) - sizeof(int);
    while(i >= 0)
      {
	*(int*)(dest + i) = *(int*)(src + i);
	i -= sizeof(int);
      }
  }
#else
  bcopy(src, dest, USE_MACHINE(pagesize));
#endif
  
  Cilk_exit_state(STATE_COPY_PAGE);
}

static inline void make_twin(void *page, void *twin)
{
  int *t;
  int i;
  
  Cilk_enter_state(STATE_TWIN_PAGE);
  
  t = twin;
  i = USE_MACHINE(pagesize) - sizeof(int);
  while(i >= 0)
    {
      *t = *(int*)(page + i);
      t++;
      i -= sizeof(int);
    }
  
  Cilk_exit_state(STATE_TWIN_PAGE);
}

static inline void apply_data(void *data, void *page, void *twin)
{
  int *t;
  int datum;
  int i;
  
  /* Cilk_dprintf("applying data %p to page %p (twin %p)\n", data, page, twin); */
  
  t = twin;
  i = USE_MACHINE(pagesize) - sizeof(int);
  while(i >= 0)
    {
      datum = *(int*)(data + i);
      if (datum != *t)
	{
	  *(int*)(page + i) = *t = datum;
	  /* we could do these assignments with SPARC v9 conditional stores! */
	}
      t++;
      i -= sizeof(int);
    }
}

static inline void apply_zero_data(void *page, void *twin)
{
  int *t;
  int datum;
  int i;
  
  t = twin;
  i = USE_MACHINE(pagesize) - sizeof(int);
  while(i >= 0)
    {
      datum = 0;
      if (datum != *t)
	{
	  *(int*)(page + i) = *t = datum;
	  /* we could do these assignments with SPARC v9 conditional stores! */
	}
      t++;
      i -= sizeof(int);
    }
}

static inline int compute_diff(void *page, void *twin, RecEntry *diff)
{
  int *t;
  int datum;
  int i;
  RecEntry *entry;
  int size;
  
  Cilk_enter_state(STATE_COMPUTE_DIFF);
  
  entry = diff;
  t = twin;
  i = USE_MACHINE(pagesize) - sizeof(int);
  
  while(i >= 0)
    {
      datum = *(int*)(page + i);
      if (datum != *t)
	{
	  *t = datum;
	  entry->offset = i;
	  entry->val = datum;
	  entry++;
	}
      t++;
      i -= sizeof(int);
    }
  size = entry - diff;
  
  Cilk_exit_state(STATE_COMPUTE_DIFF);
  
  return size;
}

static inline void apply_diff(RecEntry *diff, int size, void *page)
{
  RecEntry *end;
  
  Cilk_enter_state(STATE_APPLY_DIFF);
  
  end = diff + size;
  for(; diff < end; diff++)
    {
      *(int*)(page + diff->offset) = diff->val;
    }
  
  Cilk_exit_state(STATE_APPLY_DIFF);
}

/***********************************************************\
 * Cache interface
\***********************************************************/
void Cilk_cache_init(const Options *options);

void Cilk_reconcile_cache(void);
void Cilk_reconcile_cache_from(int proc);

void Cilk_cache_acquire_local_from(int proc);
void Cilk_cache_acquire_local(void);
void Cilk_flush_cache(void);          /* acquire global */

void Cilk_cache_fetch_reply(FetchReply *reply);
void Cilk_cache_reconcile_reply(ReconcileReply *reply);

/***********************************************************\
 * Backing store interface
\***********************************************************/
void Cilk_bstore_init(void);
void Cilk_bstore_done(void);
void Cilk_bstore_fetch_req(FetchRequest *req);
void Cilk_bstore_reconcile_req(ReconcileRequest *req);

#endif				/* _SHMEM_DISTRIB_H */
